#include <string>
#include <sstream>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
using namespace std;
#include <Ice/Ice.h>
#include <IceSSL/Plugin.h>
#include <boost/python.hpp>
#include <boost/python/operators.hpp>
#include <boost/python/class.hpp>
#include <boost/python/module.hpp>
#include <boost/python/def.hpp>

using namespace boost::python;


struct CommunicatorObject
{
    PyObject_HEAD
    Ice::CommunicatorPtr* communicator;
    /* ignored fields:
    PyObject* wrapper;
    IceUtil::Monitor<IceUtil::Mutex>* shutdownMonitor;
    WaitForShutdownThreadPtr* shutdownThread;
    bool shutdown;
    */
};


void* extract_ice_communicator(PyObject *obj)
{
    static object type;

    try
    {
        if (!obj)
            return 0;

        if (type.ptr() == Py_None)
        {
            // to enshure type safety we get system dictionary of modules
            // (AKA as sys.modules), from it we get a module 'IcePy',
            // and it's attrebute 'Connection'. It is the type
            // we check given object against.
            // This code does not have races, because (a) Python has global
            // lock (b) nothing wrong could happen in case it didn't
            object modules(handle<>(
                        borrowed(PyImport_GetModuleDict())));

            object IcePy = modules.attr("get")("IcePy");
            if (IcePy.ptr() != Py_None)
                type = IcePy.attr("Communicator");
        }

        PyTypeObject *tobj = (PyTypeObject*) type.ptr();
        if (PyObject_TypeCheck(obj, tobj))
        {
            return static_cast<void*>(
                    reinterpret_cast<CommunicatorObject*>(obj)->communicator
                    );
        }

        object arg(handle<>(borrowed(obj)));
        object impl = arg.attr("_impl");
        if ( PyObject_TypeCheck(impl.ptr(), tobj) )
        {
            return static_cast<void*>(
                    reinterpret_cast<CommunicatorObject*>(impl.ptr())->communicator
                    );
        }
    }
    catch(...)
    {}

    return 0;
}

struct ConnectionObject
{
    PyObject_HEAD
    Ice::ConnectionPtr* connection;
    Ice::CommunicatorPtr* communicator;
};


void* extract_ice_connection(PyObject *obj)
{
    static object type;

    try
    {
        if (!obj)
            return 0;

        if (type.ptr() == Py_None)
        {
            // to enshure type safety we get system dictionary of modules
            // (AKA as sys.modules), from it we get a module 'IcePy',
            // and it's attrebute 'Connection'. It is the type
            // we check given object against.
            // This code does not have races, because (a) Python has global
            // lock (b) nothing wrong could happen in case it didn't
            object modules(handle<>(
                        borrowed(PyImport_GetModuleDict())));

            type = modules["IcePy"].attr("Connection");
        }

        PyTypeObject *tobj = (PyTypeObject*) type.ptr();
        if (!PyObject_TypeCheck(obj, tobj))
            return 0;

        return static_cast<void*>(
                reinterpret_cast<ConnectionObject*>(obj)->connection
                );
    }
    catch(...)
    {}

    return 0;
}


struct ConnectionInfo
{
    //
    // The name of the negotiated cipher.
    //
    std::string cipher;

    //
    // The name of authenticated kerberos principal.
    //
    std::string krb5Princ;

    //
    // The local TCP/IP host.
    //
    std::string localAddr;

    //
    // The local port.
    //
    unsigned int localPort;

    //
    // The remote TCP/IP host.
    //
    std::string remoteAddr;

    //
    // The remote port.
    //
    unsigned int remotePort;

    //
    // If the connection is incoming this bool is true, false
    // otherwise.
    //
    bool incoming;

    //
    // The name of the object adapter that hosts this endpoint, if
    // any.
    //
    std::string adapterName;
};

//addrToString
std::string addrToString(const struct sockaddr* ad, size_t size)
{
    char namebuf[1024] = {0,};
    getnameinfo(ad, size, namebuf, sizeof(namebuf), 0,0,
            NI_NUMERICHOST | NI_NUMERICSERV);
    return std::string(namebuf);
}

static inline
std::string addrToString(const struct sockaddr_in* ad, size_t size)
{
    return addrToString(reinterpret_cast<const struct sockaddr *>(ad), size);
}

static inline
std::string addrToString(const struct sockaddr_storage* ad, size_t size)
{
    return addrToString(reinterpret_cast<const struct sockaddr *>(ad), size);
}

//addrToPort
static inline
unsigned int addrToPort(const struct sockaddr* ad)
{
    if (ad->sa_family == AF_INET)
    {
        return ntohs(reinterpret_cast<const sockaddr_in*>(ad)->sin_port);
    }
    else
    {
        return ntohs(reinterpret_cast<const sockaddr_in6*>(ad)->sin6_port);
    }
}

static inline
unsigned int addrToPort(const struct sockaddr_in* ad)
{
    return addrToPort(reinterpret_cast<const struct sockaddr *>(ad));
}

static inline
unsigned int addrToPort(const struct sockaddr_storage* ad)
{
    return addrToPort(reinterpret_cast<const struct sockaddr *>(ad));
}


ConnectionInfo makeConnectionInfo(const IceSSL::ConnectionInfo& info)
{
    ConnectionInfo res;

    res.cipher = info.cipher;
    res.krb5Princ = info.krb5Princ;
    res.localAddr = addrToString(&info.localAddr, sizeof(info.localAddr));
    res.localPort = addrToPort(&info.localAddr);
    res.remoteAddr = addrToString(&info.remoteAddr, sizeof(info.remoteAddr));
    res.remotePort = addrToPort(&info.remoteAddr);
    res.incoming = info.incoming;
    res.adapterName = info.adapterName;

    return res;
}

ConnectionInfo getConnectionInfo(Ice::ConnectionPtr* conn)
{
    return makeConnectionInfo(IceSSL::getConnectionInfo(*conn));
}

std::ostream& operator<<(std::ostream& os, ConnectionInfo const& con)
{
    os  << "{ "
        << "cipher: " << con.cipher << ", "
        << "krb5Princ: " << con.krb5Princ << ", "
        << "localAddr: " << con.localAddr << ", "
        << "localPort: " << con.localPort << ", "
        << "remoteAddr: " << con.remoteAddr << ", "
        << "remotePort: " << con.remotePort << ", "
        << "incoming: " << ( con.incoming ? "True" : "False" ) << ", "
        << "adapterName: " << con.adapterName 
        << " }";
    return os;
}


class PythonicVerifier : public IceSSL::CertificateVerifier
{
    object verify_;

public:
    PythonicVerifier(object call_verify)
    : verify_(call_verify)
    {}

    bool verify(const IceSSL::ConnectionInfo& info)
    {
        bool res = false;
        PyGILState_STATE GILState = PyGILState_Ensure();

        try
        {
            ConnectionInfo ci = makeConnectionInfo(info);
            res = extract<bool>(verify_(ci));
        }
        catch (const error_already_set&)
        {
            PyErr_Print();
            PyGILState_Release(GILState);
            return false;
        }
        catch (...)
        {
            PyGILState_Release(GILState);
            return false;
        }

        PyGILState_Release(GILState);
        return res;
    }
};

void setCertificateVerifier(Ice::CommunicatorPtr *pcomm, object verify)
{
    Ice::CommunicatorPtr comm = *pcomm;
    Ice::PluginManagerPtr pluginMgr = comm->getPluginManager();
    Ice::PluginPtr plugin = pluginMgr->getPlugin("IceSSL");
    IceSSL::PluginPtr sslPlugin = IceSSL::PluginPtr::dynamicCast(plugin);

    IceSSL::CertificateVerifierPtr cv(new PythonicVerifier(verify));
    sslPlugin->setCertificateVerifier(cv);
}


void initializePlugins(Ice::CommunicatorPtr *pcomm)
{
    Ice::CommunicatorPtr comm = *pcomm;
    Ice::PluginManagerPtr pluginMgr = comm->getPluginManager();
    pluginMgr->initializePlugins();
}


// export stuff
void cie_translator(const IceSSL::ConnectionInvalidException& ex)
{
    PyErr_SetString(PyExc_TypeError, const_cast<char*>(ex.reason.c_str()));
}


BOOST_PYTHON_MODULE(IceSSL)
{
    boost::python::converter::registry::insert(
            &extract_ice_communicator,
            type_id<Ice::CommunicatorPtr>()
            );

    boost::python::converter::registry::insert(
            &extract_ice_connection,
            type_id<Ice::ConnectionPtr>()
            );

    register_exception_translator<
        IceSSL::ConnectionInvalidException
        >(cie_translator);

    class_<ConnectionInfo>("ConnectionInfo")
        .def(self_ns::str(self))
        .def_readonly("cipher", &ConnectionInfo::cipher)
        .def_readonly("krb5Princ", &ConnectionInfo::krb5Princ)
        .def_readonly("localAddr", &ConnectionInfo::localAddr)
        .def_readonly("localPort", &ConnectionInfo::localPort)
        .def_readonly("remoteAddr", &ConnectionInfo::remoteAddr)
        .def_readonly("remotePort", &ConnectionInfo::remotePort)
        .def_readonly("incoming", &ConnectionInfo::incoming)
        .def_readonly("adapterName", &ConnectionInfo::adapterName)
        ;

    def("getConnectionInfo", getConnectionInfo);
    def("setCertificateVerifier", setCertificateVerifier);
    def("initializePlugins", initializePlugins);
}

